/******************************************************************************
*
* shell.c - implementation of "MicroStamp11 GNU-C runtime" shell
*
* Copyright (c) 2002, 2004, Don Woodward.
*
*
* This file is part of "MicroStamp11 GNU-C runtime".
*
* "MicroStamp11 GNU-C runtime" is free software; you can redistribute
* it and/or modify it under the terms of the GNU General Public License
* as published by the Free Software Foundation; either version 2 of the
* License, or any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
******************************************************************************/

#include "microstamp11.h"
#include "shell.h"

/******************************************************************************
* shell.c private typedefs
******************************************************************************/
typedef void (Shell_Func)(const char * pCommandLine);

/******************************************************************************
* shell.c private function forward declarations
******************************************************************************/
static void Shell_Help(const char * pCommandLine);
static void Shell_ShowStack(const char * pCommandLine);
static void Shell_SleepTest(const char * pCommandLine);
static void Shell_TypeInfo(const char * pCommandLine);
static void Shell_Uptime(const char * pCommandLine);

/******************************************************************************
* shell.c private static variables
******************************************************************************/
typedef struct _ShellCommand
{
    const char * m_pCommand;
    Shell_Func * m_pShellFunc;
    const char * m_pHelpText;
} ShellCommand;

/* NOTE: Can't have initialized static variables because they go into the data
   section which will overwrite the MicroStamp code loader.
   Instead make sure all statics are either uninitialized (.bss section) or
   static CONST so they go into .rodata.
*/
static const ShellCommand s_shellCommands[] =
{
    /* make sure each arg is no more than 7 chars: see Shell() */
    { "sleep",      Shell_SleepTest,    "Demonstrates use of sleep() function" },  
    { "stack",      Shell_ShowStack,    "Displays stack usage statistics" },  
    { "types",      Shell_TypeInfo,     "Displays info on C types" },  
    { "uptime",     Shell_Uptime,       "Displays time in seconds and ms since boot" },  
    { "help",       Shell_Help,         "Displays this list" },
    { "?",          Shell_Help,         "Displays this list" },
    { 0,      0,            0 }
};

/******************************************************************************
* shell.c extern constants
******************************************************************************/
extern const u16 __data_section_start;
extern const u16 __data_section_size;
extern const u16 _edata;
extern const u16 __bss_size;
extern const u16 _stack;

/******************************************************************************
* shell.c private static constants
******************************************************************************/
static const u16 s_kDataSectionStart = (u16)&__data_section_start;
static const u16 s_kDataSectionSize = (u16)&__data_section_size;
static const u16 s_kBssSectionStart = (u16)&_edata;
static const u16 s_kBssSectionSize = (u16)&__bss_size;
static const u16 s_kStackStart = (u16)&_stack;

/******************************************************************************
* shell.c private static helper functions
******************************************************************************/
static boolean
GetCommandLineArg(const char * pCommandLine, u8 nArg, char * pBuff, u16 nBuffLen)
{
    char * pCurr;

    if ((NULL == pCommandLine)  || (NULL == pBuff)|| (nBuffLen < 2)) return FALSE;
    
    /* prepare for failure */
    *pBuff = 0;

    while (nArg)
    {
        while ((*pCommandLine) && (*pCommandLine != ' ')) pCommandLine++;
        while (*pCommandLine == ' ') pCommandLine++;
        nArg--;
    }

    while (*pCommandLine == ' ') pCommandLine++;
    if (*pCommandLine == 0) return FALSE;

    pCurr = pBuff;
    while ((nBuffLen > 1) && (*pCommandLine) && (*pCommandLine != ' '))
    {
        *pCurr++ = *pCommandLine++;
        nBuffLen--;  
    }

    /* see if we finished */
    if ((*pCommandLine == 0) || (*pCommandLine == ' '))
    {
        *pCurr = 0;
        return TRUE;
    }
    else return FALSE;
}

static void
CopyOverlappingString(char * s1, char * s2)
{
    if ((s1 >= s2) || (NULL == s1) || (NULL == s2)) return;
    while (*s2) *s1++ = *s2++;
    *s1 = 0;
}

static void
NumberToStringDecimal32(u32 nNumber, char * pBuff, u16 nBuffLen)
{
    char * pChar;
    u16 remainder;

    if ((NULL == pBuff) || (nBuffLen < 2)) return;
    
    if (nNumber == 0)
    {
        *pBuff++ = '0';
        *pBuff = 0;
        return;
    }

    pChar = pBuff + nBuffLen - 1;
    *pChar-- = 0;

    while (nNumber)
    {
        if (pChar < pBuff)
        {
            /* not enough room in pBuff */
            *pBuff = 0;
            return;
        }

        remainder = nNumber % 10; /* damn, this MOD is going to generate code!!! */
        *pChar-- = remainder + '0';
        nNumber /= 10;
    }

    CopyOverlappingString(pBuff, pChar + 1);
    return;
}

static void
NumberToStringDecimal(u16 nNumber, char * pBuff, u8 nBuffLen)
{
    char * pChar;
    u16 remainder;

    if ((NULL == pBuff) || (nBuffLen < 2)) return;
    
    if (nNumber == 0)
    {
        *pBuff++ = '0';
        *pBuff = 0;
        return;
    }

    pChar = pBuff + nBuffLen - 1;
    *pChar-- = 0;

    while (nNumber)
    {
        if (pChar < pBuff)
        {
            /* not enough room in pBuff */
            *pBuff = 0;
            return;
        }

        remainder = nNumber % 10; /* damn, this MOD is going to generate code!!! */
        *pChar-- = remainder + '0';
        nNumber /= 10;
    }

    CopyOverlappingString(pBuff, pChar + 1);
    return;
}

static void
NumberToStringHex(u16 nNumber, char * pBuff, u8 nBuffLen)
{
    char * pChar;
    u16 remainder;
    const char * pHex = "ABCDEF";

    if ((NULL == pBuff) || (nBuffLen < 2)) return;
    
    if (nNumber == 0)
    {
        *pBuff++ = '0';
        *pBuff = 0;
        return;
    }

    pChar = pBuff + nBuffLen - 1;
    *pChar-- = 0;

    while (nNumber)
    {
        if (pChar < pBuff)
        {
            /* not enough room in pBuff */
            *pBuff = 0;
            return;
        }

        remainder = nNumber % 16; /* damn, this MOD is going to generate code!!! */
        if (remainder < 10) *pChar-- = remainder + '0';
        else *pChar-- = pHex[remainder - 10];
        nNumber /= 16;
    }

    CopyOverlappingString(pBuff, pChar + 1);
    return;
}

static boolean
StringMatch(const char * pString1, const char * pString2)
{
    while (*pString1 == *pString2)
    {
        if (0 == *pString1) return TRUE;
        pString1++; pString2++;
    }
    return FALSE;
}

const ShellCommand *
LookupCommand(const char * pCommand)
{
    u8 nIndex = 0;  /* only 255 commands! */
    while (s_shellCommands[nIndex].m_pCommand)
    {
        if (StringMatch(pCommand, s_shellCommands[nIndex].m_pCommand)) return &s_shellCommands[nIndex];
        nIndex++;
    }
    return NULL;
}

static void
DispatchShellCommand(const char * pCommandLine)
{
    char command[8];

    if ((NULL == pCommandLine) || (0 == *pCommandLine)) return;
    if (! GetCommandLineArg(pCommandLine, 0, command, sizeof(command))) return;

    if (NULL == LookupCommand(command))
    {
        serialPutString("MicroStamp11: ");
        serialPutString(command);
        serialPutString(": Unknown command.  ? for help.\n\n");
    }
    else
    {
        (*(LookupCommand(command)->m_pShellFunc))(pCommandLine);
        serialPutChar('\n');
    }
}

/******************************************************************************
* shell.c main shell function
******************************************************************************/
void
Shell()
{
    char buff[8];

    serialPutString("\n\n\nWelcome to the MicroStamp11 Shell!\n\n");
    for (;;)
    {
        serialPutString("MicroStamp11> ");
        serialGetLine(buff, sizeof(buff), TRUE);
        DispatchShellCommand(buff);
    }
}

/******************************************************************************
* shell.c shell command functions
******************************************************************************/
static void
Shell_Help(const char * pCommandLine)
{
    u8 nIndex = 0;  /* only 255 commands! */
    u8 nColumn;
    u8 nLen;
    const char * pTableCommand;

    serialPutString("\nMicroStamp11 Shell Commands\n");
    serialPutString(  "---------------------------\n");

    while (NULL != (pTableCommand = s_shellCommands[nIndex].m_pCommand))
    {
        serialPutString(pTableCommand);
        nLen = 0;
        while (*pTableCommand++) nLen++;;
        nColumn = 14 - nLen;
        while (nColumn > 0) { serialPutChar(' '); nColumn--; }
        serialPutString("- ");
        serialPutString(s_shellCommands[nIndex].m_pHelpText);
        serialPutChar('\n');
        nIndex++;
    }
}

static void
DumpDataSection()
{
    char buff[6];

    /* Section .data:   Start: 40,  End: 50,     Size: 10 (16 bytes)        */

    serialPutString("Section .data:   Start: ");
    NumberToStringHex(s_kDataSectionStart, buff, sizeof(buff)); serialPutString(buff);
    serialPutString(",  End: ");
    NumberToStringHex(((s_kDataSectionSize == 0) ? s_kDataSectionStart : s_kDataSectionStart + s_kDataSectionSize - 1), buff, sizeof(buff)); serialPutString(buff);
    serialPutString(",     Size: ");
    NumberToStringHex(s_kDataSectionSize, buff, sizeof(buff)); serialPutString(buff);
    serialPutString(" (");
    NumberToStringDecimal(s_kDataSectionSize, buff, sizeof(buff)); serialPutString(buff);
    serialPutString(" bytes)\n");
}

static void
DumpBSSSection()
{
    char buff[6];

    /* Section .bss :   Start: 51,  End: 52,     Size: 2 (2 bytes)        */
    serialPutString("Section .bss :   Start: ");
    NumberToStringHex(s_kBssSectionStart, buff, sizeof(buff)); serialPutString(buff);
    serialPutString(",  End: ");
    NumberToStringHex(((s_kBssSectionSize == 0) ? s_kBssSectionStart : s_kBssSectionStart + s_kBssSectionSize - 1), buff, sizeof(buff)); serialPutString(buff);
    serialPutString(",     Size: ");
    NumberToStringHex(s_kBssSectionSize, buff, sizeof(buff)); serialPutString(buff);
    serialPutString(" (");
    NumberToStringDecimal(s_kBssSectionSize, buff, sizeof(buff)); serialPutString(buff);
    serialPutString(" bytes)\n");
}

static void
DumpStack()
{
    const u16 stackEnd = s_kBssSectionStart + s_kBssSectionSize;
    const u16 stackSize = s_kStackStart - stackEnd;
    char buff[6];

    /*         Stack:  Bottom: FF,  Top: 50,     Size: B0 (176 bytes)       */
    
    serialPutString("        Stack:  Bottom: ");
    NumberToStringHex(s_kStackStart, buff, sizeof(buff)); serialPutString(buff);
    serialPutString(",  Top: ");
    NumberToStringHex(stackEnd, buff, sizeof(buff)); serialPutString(buff);
    serialPutString(",     Size: ");
    NumberToStringHex(stackSize, buff, sizeof(buff)); serialPutString(buff);
    serialPutString(" (");
    NumberToStringDecimal(stackSize, buff, sizeof(buff)); serialPutString(buff);
    serialPutString(" bytes)\n");
}

static void
DumpEmptyStack()
{
    const u16 stackEnd = s_kBssSectionStart + s_kBssSectionSize;
    const u16 stackSize = s_kStackStart - stackEnd;
    char buff[6];

    /*   Empty Stack: Pointer: FF, Size: 50, Headroom: B0 (176 bytes)       */
    serialPutString("  Empty Stack: Pointer: ");
    NumberToStringHex(s_kStackStart, buff, sizeof(buff)); serialPutString(buff);
    serialPutString(", Size: ");
    NumberToStringHex(stackSize, buff, sizeof(buff)); serialPutString(buff);
    serialPutString(", Headroom: ");
    NumberToStringHex(stackSize, buff, sizeof(buff)); serialPutString(buff);
    serialPutString(" (");
    NumberToStringDecimal(stackSize, buff, sizeof(buff)); serialPutString(buff);
    serialPutString(" bytes)\n");
}

static void
DumpMaxStack()
{
    const u16 stackEnd = s_kBssSectionStart + s_kBssSectionSize;
    u16 currMaxStack;
    char buff[6];

    getMaxStack(&currMaxStack);

    /*     Max Stack: Pointer: FF, Size: 50, Headroom: B0 (176 bytes)       */

    serialPutString("    Max Stack: Pointer: ");
    NumberToStringHex(currMaxStack, buff, sizeof(buff)); serialPutString(buff);
    serialPutString(", Size: ");
    NumberToStringHex((s_kStackStart - currMaxStack), buff, sizeof(buff)); serialPutString(buff);
    serialPutString(", Headroom: ");
    NumberToStringHex(currMaxStack - stackEnd + 1, buff, sizeof(buff)); serialPutString(buff);
    serialPutString(" (");
    NumberToStringDecimal(currMaxStack - stackEnd + 1, buff, sizeof(buff)); serialPutString(buff);
    serialPutString(" bytes)\n");
}

static void
Shell_ShowStack(const char * pCommandLine)
{
    serialPutChar('\n');
    DumpDataSection();
    DumpBSSSection();
    DumpStack();
    DumpEmptyStack();
    DumpMaxStack();
}

static void
Shell_SleepTest(const char * pCommandLine)
{
    u8 i;
    char buff[6];
    for (i = 0; i <= 10; i++)
    {
        serialPutString("sleep(");
        if (i == 0) serialPutString("   ");
        else if (i < 10) serialPutString(" ");
        NumberToStringDecimal(i * 100, buff, sizeof(buff)); 
        serialPutString(buff);
        serialPutString(")... ");
        sleep(i * 100);
        serialPutString("Awake!\n");
    }

    serialPutString("\nDone!");
}

static void
Shell_Uptime(const char * pCommandLine)
{
    TimeStruct timeStruct;
    char buff[8];

    getKernelTime(&timeStruct);
    serialPutString("\nUptime: ");
    NumberToStringDecimal32(timeStruct.m_nSeconds, buff, sizeof(buff));
    serialPutString(buff);
    serialPutString(" seconds, ");
    NumberToStringDecimal(timeStruct.m_nMilliseconds, buff, sizeof(buff));
    serialPutString(buff);
    serialPutString(" milliseconds.\n");
}

static void
Shell_TypeInfo(const char * pCommandLine)
{
    char buff[4];

    serialPutString("sizeof(u8)      = "); NumberToStringDecimal(sizeof(u8),       buff, sizeof(buff)); serialPutString(buff); serialPutChar('\n');
    serialPutString("sizeof(u16)     = "); NumberToStringDecimal(sizeof(u16),      buff, sizeof(buff)); serialPutString(buff); serialPutChar('\n');
    serialPutString("sizeof(u32)     = "); NumberToStringDecimal(sizeof(u32),      buff, sizeof(buff)); serialPutString(buff); serialPutChar('\n');
    serialPutString("sizeof(s8)      = "); NumberToStringDecimal(sizeof(s8),       buff, sizeof(buff)); serialPutString(buff); serialPutChar('\n');
    serialPutString("sizeof(s16)     = "); NumberToStringDecimal(sizeof(s16),      buff, sizeof(buff)); serialPutString(buff); serialPutChar('\n');
    serialPutString("sizeof(s32)     = "); NumberToStringDecimal(sizeof(s32),      buff, sizeof(buff)); serialPutString(buff); serialPutChar('\n');
    serialPutString("sizeof(boolean) = "); NumberToStringDecimal(sizeof(boolean),  buff, sizeof(buff)); serialPutString(buff); serialPutChar('\n');
    serialPutString("sizeof(char)    = "); NumberToStringDecimal(sizeof(char),     buff, sizeof(buff)); serialPutString(buff); serialPutChar('\n');
    serialPutString("sizeof(int)     = "); NumberToStringDecimal(sizeof(int),      buff, sizeof(buff)); serialPutString(buff); serialPutChar('\n');
    serialPutString("sizeof(short)   = "); NumberToStringDecimal(sizeof(short),    buff, sizeof(buff)); serialPutString(buff); serialPutChar('\n');
    serialPutString("sizeof(long)    = "); NumberToStringDecimal(sizeof(long),     buff, sizeof(buff)); serialPutString(buff); serialPutChar('\n');
}


/******************************************************************************
*   LOG
*******************************************************************************
*   03-Jan-02   dwoodward   created
*
*/
